/*
    Astico2D: Another Simple Tool for Image Computing with OpenCV in 2D
    https://pageperso.lis-lab.fr/edouard.thiel/Astico2D/

    astico2d.cpp - main module 

    CC BY-SA Edouard.Thiel@univ-amu.fr - 30/07/2023 - v1.1
*/

#include "astico2d.hpp"


//------------------------------ C O L O R S ----------------------------------

void translate_to_vga_colors (cv::Mat &img_int, cv::Mat &img_col)
{
    CHECK_MAT_TYPE(img_int, CV_32SC1);
    CHECK_MAT_TYPE(img_col, CV_8UC3);

    unsigned char colors[16][3] = {  // R, G, B
        {   0,   0,   0 },   //  0  black           ->  0 only
        {  30,  30, 230 },   //  1  blue            ->  1, 15, 29, ...
        {  30, 200,  30 },   //  2  green           ->  2, 16, 30, ...
        {  30, 200, 200 },   //  3  cyan            ->  3, 17, 31, ...
        { 200,  30,  30 },   //  4  red             ->  4, 18, 32, ...
        { 200,  30, 200 },   //  5  magenta         ->  5, 19, 33, ...
        { 200, 130,  50 },   //  6  brown           ->  6, 20, 34, ...
        { 200, 200, 200 },   //  7  light gray      ->  7, 21, 35, ...
        { 110, 110, 140 },   //  8  dark gray       ->  8, 22, 36, ...
        {  84, 130, 252 },   //  9  light blue      ->  9, 23, 37, ...
        {  84, 252,  84 },   // 10  light green     -> 10, 24, 38, ...
        {  84, 252, 252 },   // 11  light cyan      -> 11, 25, 39, ...
        { 252,  84,  84 },   // 12  light red       -> 12, 26, 40, ...
        { 252,  84, 252 },   // 13  light magenta   -> 13, 27, 41, ...
        { 252, 252,  84 },   // 14  yellow          -> 14, 28, 42, ...
        { 252, 252, 252 },   // 15  white           -> 255 only
    };

    for (int y = 0; y < img_int.rows; y++)
    for (int x = 0; x < img_int.cols; x++)
    {
        int g = img_int.at<int>(y,x), c = 0;
        if (g == 255) c = 15;                      // only 255 is white
        else if (g != 0) c = 1 + abs(g-1) % 14;    // only 0 is black
        // Attention img_col is in B, G, R order -> reverse channels
        img_col.at<cv::Vec3b>(y,x)[0] = colors[c][2];
        img_col.at<cv::Vec3b>(y,x)[1] = colors[c][1];
        img_col.at<cv::Vec3b>(y,x)[2] = colors[c][0];
    }
}


void reverse_img_colors (cv::Mat &img)
{
    CHECK_MAT_TYPE(img, CV_8UC3);

    for (int y = 0; y < img.rows; y++)
    for (int x = 0; x < img.cols; x++)
    {
        img.at<cv::Vec3b>(y,x)[0] = 255 - img.at<cv::Vec3b>(y,x)[0];
        img.at<cv::Vec3b>(y,x)[1] = 255 - img.at<cv::Vec3b>(y,x)[1];
        img.at<cv::Vec3b>(y,x)[2] = 255 - img.at<cv::Vec3b>(y,x)[2];
    }
}


void reverse_pixel_colors (cv::Mat &img, int x, int y)
{
    CHECK_MAT_TYPE(img, CV_8UC3);

    if (0 <= y && y < img.rows && 0 <= x && x < img.cols)
    {
        img.at<cv::Vec3b>(y,x)[0] = 255 - img.at<cv::Vec3b>(y,x)[0];
        img.at<cv::Vec3b>(y,x)[1] = 255 - img.at<cv::Vec3b>(y,x)[1];
        img.at<cv::Vec3b>(y,x)[2] = 255 - img.at<cv::Vec3b>(y,x)[2];
    }
}


//--------------------------D R A W I N G S ---------------------------

// Use any size of patterns, e.g. {0,0,1}, {0,0,0,1,1,1}

void draw_dotted_line (cv::Mat &img, const cv::Point &p1, 
    const cv::Point &p2, const cv::Scalar &color1, const cv::Scalar &color2, 
    const std::vector<bool> &pattern)
{
    CHECK_MAT_TYPE(img, CV_8UC3);

    cv::LineIterator it (img, p1, p2, cv::LINE_8);
    for (int i = 0; i < it.count; i++, it++)
    {
        if (pattern[i % pattern.size()]) {
            (*it)[0] = color1.val[0];
            (*it)[1] = color1.val[1];
            (*it)[2] = color1.val[2];
        } else {
            (*it)[0] = color2.val[0];
            (*it)[1] = color2.val[1];
            (*it)[2] = color2.val[2];
        }
    }
}


// Usage:
//   draw_dotted_rectangle (img_col, cv::Point(10,50), cv::Point(20,70),
//       cv::Scalar(0,0,255), cv::Scalar(0,255,0), {0,0,0,1,1,1,1});

void draw_dotted_rectangle (cv::Mat &img, const cv::Point &p1, 
    const cv::Point &p2, const cv::Scalar &color1, const cv::Scalar &color2, 
    const std::vector<bool> &pattern)
{
    cv::Point p3 = cv::Point(p2.x,p1.y), p4 = cv::Point(p1.x,p2.y);
    draw_dotted_line (img, p1, p3, color1, color2, pattern);
    draw_dotted_line (img, p3, p2, color1, color2, pattern);
    draw_dotted_line (img, p2, p4, color1, color2, pattern);
    draw_dotted_line (img, p4, p1, color1, color2, pattern);
}


void draw_dotted_cross (cv::Mat &img, const cv::Point &p1, 
    int l1, int l2, const cv::Scalar &color1, const cv::Scalar &color2, 
    const std::vector<bool> &pattern)
{
    cv::Point p2 = cv::Point(p1.x, p1.y-l1),
              p3 = cv::Point(p1.x, p1.y-l2),
              p4 = cv::Point(p1.x, p1.y+l1),
              p5 = cv::Point(p1.x, p1.y+l2),
              p6 = cv::Point(p1.x-l1, p1.y),
              p7 = cv::Point(p1.x-l2, p1.y),
              p8 = cv::Point(p1.x+l1, p1.y),
              p9 = cv::Point(p1.x+l2, p1.y);
    draw_dotted_line (img, p2, p3, color1, color2, pattern);
    draw_dotted_line (img, p4, p5, color1, color2, pattern);
    draw_dotted_line (img, p6, p7, color1, color2, pattern);
    draw_dotted_line (img, p8, p9, color1, color2, pattern);
}


//---------------------------------- T E X T ----------------------------------

void AoText::compute_size (std::string s, int &w, int &h)
{
    cv::Size ts = cv::getTextSize(s, fontFace, fontScale, thickness, &baseline);
    w = ts.width;
    h = ts.height + baseline - thickness;
    offset_h = ts.height + fontScale;
}


// x,y is the top-left corner.
// Remark: the effective baseline will be at y + h + thickness - baseline

void AoText::draw_text (std::string s, int x, int y, cv::Mat &dest)
{
    cv::putText(dest, s, cv::Point (x, y+offset_h), fontFace, fontScale,
        color, thickness, lineType);
}


//---------------------------------- P L A C E --------------------------------


void AoPlace::init (int h, int w)
{
    src_h = h;
    src_w = w;
    if (zoom_h + footer_h < view_min) zoom_h = view_min - footer_h;
    if (zoom_w < view_min) zoom_w = view_min;
    adjust_bounds ();
    compute_map_scale ();
}


void AoPlace::adjust_bounds ()
{
    int dh = zoom_h / zoom;
    int dw = zoom_w / zoom;

    if (zoom_x0 < 0) zoom_x0 = 0;
    zoom_x1 = zoom_x0 + dw;
    if (zoom_x1 > src_w) {
        zoom_x1 = src_w;
        zoom_x0 = zoom_x1 - dw;
        if (zoom_x0 < 0) zoom_x0 = 0;
    }

    if (zoom_y0 < 0) zoom_y0 = 0;
    zoom_y1 = zoom_y0 + dh;
    if (zoom_y1 > src_h) {
        zoom_y1 = src_h;
        zoom_y0 = zoom_y1 - dh;
        if (zoom_y0 < 0) zoom_y0 = 0;
    }
}


void AoPlace::move (int dx, int dy)
{
    zoom_x0 += dx; zoom_y0 += dy; 
    zoom_x1 += dx; zoom_y1 += dy; 
    adjust_bounds ();
}


void AoPlace::resize_view (int dir_h, int dir_w)
{
    if (dir_h < 0) {
        zoom_h = zoom_h - view_inc;
        if (zoom_h + footer_h < view_min) zoom_h = view_min - footer_h;
    } else if (dir_h > 0) {
        zoom_h = zoom_h + view_inc;
        if (zoom_h + footer_h > view_max) zoom_h = view_max - footer_h;
    }

    if (dir_w < 0) {
        zoom_w = zoom_w - view_inc;
        if (zoom_w < view_min) zoom_w = view_min;
    } else if (dir_w > 0) {
        zoom_w = zoom_w + view_inc;
        if (zoom_w > view_max) zoom_w = view_max;
    }

    //std::cout << "View size " << zoom_w << "x" << zoom_h << std::endl;
    adjust_bounds ();
}


int AoPlace::compute_optimal_zoom ()
{
    int zh = opt_h / src_h, 
        zw = opt_w / src_w;
    int z = zh < zw ? zh : zw;
    if (z < 1) z = 1; else if (z > zoom_max) z = zoom_max;
    return z;
}


void AoPlace::resize_view_opt ()
{
    zoom_h = ((src_h * zoom - 1) / view_inc +1) * view_inc;
    if (zoom_h + footer_h < view_min) zoom_h = view_min - footer_h;
    else if (zoom_h > opt_h) zoom_h = opt_h;

    zoom_w = ((src_w * zoom - 1) / view_inc +1) * view_inc;
    if (zoom_w < view_min) zoom_w = view_min; 
    else if (zoom_w > opt_w) zoom_w = opt_w;

    //std::cout << "View size " << zoom_w << "x" << zoom_h << std::endl;
    adjust_bounds ();
}


void AoPlace::get_cell_coords (int ev_x, int ev_y, int &cx, int &cy)
{
    cx = zoom_x0 + ev_x / zoom;
    cy = zoom_y0 + ev_y / zoom;
}


void AoPlace::draw_cell_rect (cv::Mat &dest, int cx, int cy)
{
    CHECK_MAT_TYPE(dest, CV_8UC3);

    if (cx < 0 || cx >= src_w || cy < 0 || cy >= src_h) return;

    int ax = (cx - zoom_x0)*zoom,
        ay = (cy - zoom_y0)*zoom;
    cv::Point p1 = cv::Point(ax-1, ay-1),
              p2 = cv::Point(ax+zoom+1, ay+zoom+1);
    draw_dotted_rectangle (dest, p1, p2,
        cv::Scalar(0,255,255), cv::Scalar(0,0,0), {0,0,0,1,1,1,1});
}


void AoPlace::draw_cell_cross (cv::Mat &dest, int cx, int cy)
{
    CHECK_MAT_TYPE(dest, CV_8UC3);

    if (cx < 0 || cx >= src_w || cy < 0 || cy >= src_h) return;

    int ax = (cx - zoom_x0)*zoom + zoom/2,
        ay = (cy - zoom_y0)*zoom + zoom/2;
    cv::Point p1 = cv::Point(ax, ay);

    draw_dotted_cross (dest, p1, zoom/2+3, zoom/2+23,
        cv::Scalar(0,255,255), cv::Scalar(0,0,0), {0,0,0,1,1,1,1});
}


void AoPlace::draw_location (cv::Mat &src, cv::Mat &dest)
{
    CHECK_MAT_TYPE(src, CV_8UC3);
    CHECK_MAT_TYPE(dest, CV_8UC3);

    cv::resize(src, dest, cv::Size(), map_f, map_f, cv::INTER_NEAREST);
    cv::Point p0 = cv::Point(zoom_x0*map_f, zoom_y0*map_f),
              p1 = cv::Point(zoom_x1*map_f, zoom_y1*map_f);
    cv::rectangle(dest, p0, p1, cv::Scalar (255, 255, 255), 3, 4);
    cv::rectangle(dest, p0, p1, cv::Scalar (  0,   0, 255), 1, 4);
}


void AoPlace::draw_image_portion (cv::Mat &src, cv::Mat &dest)
{
    CHECK_MAT_TYPE(src, CV_8UC3);
    CHECK_MAT_TYPE(dest, CV_8UC3);

    for (int y = 0; y < dest.rows-footer_h; y++)
    for (int x = 0; x < dest.cols; x++)
    {
        int x0, y0;
        get_cell_coords (x, y, x0, y0);

        if (x0 < 0 || x0 >= src.cols || y0 < 0 || y0 >= src.rows) {
            dest.at<cv::Vec3b>(y,x)[0] = 64;
            dest.at<cv::Vec3b>(y,x)[1] = 64;
            dest.at<cv::Vec3b>(y,x)[2] = 64;
            continue;
        }
        dest.at<cv::Vec3b>(y,x)[0] = src.at<cv::Vec3b>(y0,x0)[0];
        dest.at<cv::Vec3b>(y,x)[1] = src.at<cv::Vec3b>(y0,x0)[1];
        dest.at<cv::Vec3b>(y,x)[2] = src.at<cv::Vec3b>(y0,x0)[2];
    }
}


void AoPlace::draw_footer (cv::Mat &dest)
{
    CHECK_MAT_TYPE(dest, CV_8UC3);

    cv::Point p0 = cv::Point(0, dest.rows-footer_h),
              p1 = cv::Point(dest.cols-1, dest.rows-1);
    cv::rectangle(dest, p0, p1, cv::Scalar (240, 241, 242), cv::FILLED);

    AoText t;
    int w, h, y;
    t.compute_size (msg_left_footer, w, h);
    y = dest.rows - footer_h + (footer_h - h)/2 +1;
    t.draw_text (msg_left_footer, 6, y, dest);
    t.compute_size (msg_right_footer, w, h);
    t.draw_text (msg_right_footer, dest.cols-6-w, y, dest);
}


void AoPlace::print_array_values (cv::Mat &src, int ex, int ey, int rx, int ry)
{
    CHECK_MAT_TYPE(src, CV_32SC1);

    int cx, cy;
    get_cell_coords (ex, ey, cx, cy);
    int x1 = cx-rx, x2 = cx+rx, y1 = cy-ry, y2 = cy+ry;

    int vmin = 0, vmax = 0;
    for (int y = y1; y <= y2; y++) {
        for (int x = x1; x <= x2; x++) {
            if (x < 0 || x >= src.cols || y < 0 || y >= src.rows)
                continue;
            int v = src.at<int>(y,x);
            if (v < vmin) vmin = v;
            else if (v > vmax) vmax = v;
        }
    }
    int n1 = ceil(log10(abs(vmin)+1)), n2 = ceil(log10(abs(vmax)+1));
    if (vmin < 0) n1++; // sign 
    if (vmax < 0) n2++;
    int n = std::max(n1,n2); // digits number

    std::cout << "Values arround " << cx << "," << cy << " :" << std::endl;

    for (int y = y1; y <= y2; y++) {
        for (int x = x1; x <= x2; x++) {
            if (x < 0 || x >= src.cols || y < 0 || y >= src.rows)
                 std::cout << std::setw(n+1) << "* " ;  // outside image
            else std::cout << std::setw(n) << src.at<int>(y,x) << " ";
        }
        std::cout << std::endl;
    }
    std::cout << std::endl;
}


void AoPlace::print_array_colors (cv::Mat &src, int ex, int ey, int rx, int ry)
{
    CHECK_MAT_TYPE(src, CV_8UC3);

    int cx, cy;
    get_cell_coords (ex, ey, cx, cy);
    int x1 = cx-rx, x2 = cx+rx, y1 = cy-ry, y2 = cy+ry;

    int n = 3; // digits number

    std::cout << "Colors arround " << cx << "," << cy << " :" << std::endl;

    for (int y = y1; y <= y2; y++) {
        for (int x = x1; x <= x2; x++) {
            if (x < 0 || x >= src.cols || y < 0 || y >= src.rows)
                // outside image
                std::cout << "("
                    << std::setw(n) << "* " << ","
                    << std::setw(n) << "* " << ","
                    << std::setw(n) << "* " << ") ";
            else {
                int cR = src.at<cv::Vec3b>(y,x)[2],
                    cG = src.at<cv::Vec3b>(y,x)[1],
                    cB = src.at<cv::Vec3b>(y,x)[0];
                std::cout << "("
                    << std::setw(n) << cR << ","
                    << std::setw(n) << cG << ","
                    << std::setw(n) << cB << ") ";
            }
        }
        std::cout << std::endl;
    }
    std::cout << std::endl;
}


void AoPlace::compute_map_scale ()
{
    map_f = (double) map_w / src_w;
}


void AoPlace::change_map_width (int dir)
{
    if (dir < 0) {
        map_w -= map_w_inc;
        if (map_w < map_w_min) map_w = map_w_min;
    } else {
        map_w += map_w_inc;
        if (map_w > map_w_max) map_w = map_w_max;
    }
    compute_map_scale();
}


//----------------------------- A S T I C O 2 D -------------------------------


Astico2D::Astico2D (int &argc, char **&argv) 
    : measure_time()
{
    if (analyse_args (argc, argv) != 0) { init_ok = false ; return; }
    if (read_input_image () != 0) { init_ok = false ; return; }
    create_internal_images ();
    create_windows ();
}


int Astico2D::run ()
{
    if (! init_ok) return 1;
    print_keyboard_keys ();
    handle_events ();
    return 0;
}


void Astico2D::print_usage () {
    std::cout << "Usage : " << prog_name
        << " [-mag w h] [-thr threshold] [-o image_res] image_src"
        << " [other options]"
        << std::endl;
}


int Astico2D::analyse_args (int &argc, char **&argv)
{
    prog_name = argv[0];
    argc -= 1; argv += 1;
    name_out2 = "tmp-output.png";

    while (argc > 0) {
        if (!strcmp(argv[0], "-mag")) {
            if (argc-1 < 3) { print_usage(); return 1; }
            place.zoom_w = atoi(argv[1]);
            place.zoom_h = atoi(argv[2]);
            argc -= 3; argv += 3;
        } else if (!strcmp(argv[0], "-thr")) {
            if (argc-1 < 2) { print_usage(); return 1; }
            thresh_val = atoi(argv[1]);
            argc -= 2; argv += 2;
        } else if (!strcmp(argv[0], "-o")) {
            if (argc-1 < 2) { print_usage(); return 1; }
            name_out2 = argv[1];
            argc -= 2; argv += 2;
        } else break;
    }

    if (argc < 1) { print_usage(); return 1; }
    name_in1  = argv[0];
    argc -= 1; argv += 1;

    // Additional arguments remain in argc,argv and can be analysed in the
    // constructor of a child class after the call to the constructor of 
    // the parent class.
    return 0;
}


int Astico2D::read_input_image ()
{
    img_src = cv::imread (name_in1, cv::IMREAD_COLOR);  // produces 8UC3
    if (img_src.empty()) {
        std::cout << "Error while reading \"" << name_in1 << "\" " 
            << "or incompatible image format." << std::endl;
        return -1;
    } else {
        std::cout << "Reading image \"" << name_in1 << "\" : " <<
            img_src.rows << "x" << img_src.cols << std::endl;
    }
    return 0;
}


int Astico2D::save_result_image ()
{
    if (name_out2 == "") return 0;

    int res = cv::imwrite (name_out2, img_col);
    if (res) std::cout << "Image successfully saved";
    else std::cout << "Error while saving image";
    std::cout << " in \"" << name_out2 << "\"." << std::endl;
    return res ? 0 : -1;
}


void Astico2D::create_internal_images ()
{
    place.init (img_src.rows, img_src.cols);
    img_res1 = cv::Mat(img_src.rows*place.map_f, 
                       img_src.cols*place.map_f, CV_8UC3);
    img_res2 = cv::Mat(place.zoom_h+place.footer_h, place.zoom_w, CV_8UC3);
    img_int  = cv::Mat(img_src.rows, img_src.cols, CV_32SC1);
    img_col = cv::Mat(img_src.rows, img_src.cols, CV_8UC3);
}


void Astico2D::create_slider (const char* slider_name, int *current_value,
    int max_value)
{
    cv::createTrackbar (slider_name, "Map", current_value, max_value, 
        onDefaultTrackbarSlide, this);
}


void Astico2D::modify_slider (const char* slider_name, int new_value)
{
    cv::setTrackbarPos (slider_name, "Map", new_value);
}


void Astico2D::decrease_slider (const char* slider_name, int min_value)
{
    int val = cv::getTrackbarPos (slider_name, "Map") - 1;
    if (val >= min_value) modify_slider (slider_name, val);
}


void Astico2D::increase_slider (const char* slider_name, int max_value)
{
    int val = cv::getTrackbarPos (slider_name, "Map") + 1;
    if (val <= max_value) modify_slider (slider_name, val);
}


void Astico2D::create_windows ()
{
    cv::namedWindow ("Map", cv::WINDOW_AUTOSIZE);
    cv::createTrackbar ("Zoom", "Map", &zoom_slider, place.zoom_max, 
        onZoomSlide, this);
    create_slider ("Thresh", &thresh_val, thresh_max);
    cv::setMouseCallback ("Map", onMouseEventMap, this);

    cv::namedWindow ("View", cv::WINDOW_AUTOSIZE);
    cv::setMouseCallback ("View", onMouseEventView, this);
}


void Astico2D::rename_view_window (std::string name)
{
    cv::setWindowTitle("View", name);
}


void Astico2D::rename_map_window (std::string name)
{
    cv::setWindowTitle("Map", name);
}


void Astico2D::set_left_footer (std::string msg)
{
    place.msg_left_footer = msg;
    raise_recomp (R_VIEW);
}


void Astico2D::set_right_footer (std::string msg)
{
    place.msg_right_footer = msg;
    raise_recomp (R_VIEW);
}


std::string Astico2D::compose_msg_map_coords ()
{
    int x = cell_x, y = cell_y;
    if (x < 0 || x >= place.src_w || y < 0 || y >= place.src_h) return "";

    std::string s = std::to_string(x)+","+std::to_string(y) + " : ";

    if (mode_transform == M_LEVELS) {
        s += std::to_string(img_int.at<int>(y,x));
    } else {
        int cR = img_col.at<cv::Vec3b>(y,x)[2],
            cG = img_col.at<cv::Vec3b>(y,x)[1],
            cB = img_col.at<cv::Vec3b>(y,x)[0];
        s += std::to_string(cR)+","+std::to_string(cG)+","+std::to_string(cB);
    }
    return s;
}


void Astico2D::raise_recomp (RecomputeImages new_recomp)
{
    if (new_recomp > recomp) recomp = new_recomp;
}


void Astico2D::handle_events ()
{
    for (;;) {

        if (recomp == R_ALL)
        {
            if (mode_transform == M_LEVELS) {
                //std::cout << "Compute img_int" << std::endl;
                cv::Mat img_gray;
                cv::cvtColor (img_src, img_gray, cv::COLOR_BGR2GRAY);
                cv::threshold (img_gray, img_gray, thresh_val, 255, 
                    cv::THRESH_BINARY);
                img_gray.convertTo (img_int, CV_32SC1,1., 0.);
            } else {
                img_col = img_src.clone();
            }

            if (mode_transform != M_NONE && key_transform != 0) {
                //std::cout << "Compute tranformations for '" 
                //    << char(key_transform) << "'" << std::endl;
                double t1 = measure_time.get_time_dbl ();
                // Note: this call must produce a color image in img_col.
                process_transformations ();
                double t2 = measure_time.get_time_dbl ();
                std::string msg = "'"+std::string{key_transform}+"'";
                if (MEASURE_TIME) {
                    //std::cout << "Duration of transformations: " 
                    //    << std::to_string(t2-t1) << " s" << std::endl;
                    msg += " : "+std::to_string(t2-t1)+"s";
                }
                set_left_footer (msg);
            } else if (mode_transform == M_LEVELS) {
                translate_to_vga_colors (img_int, img_col);
            }
        }

        if (recomp == R_ALL || recomp == R_MAP) {
            //std::cout << "Compute map image" << std::endl;
            place.draw_location (img_col, img_res1);
            cv::imshow ("Map", img_res1);
        }

        if (recomp == R_ALL || recomp == R_MAP || recomp == R_VIEW) {
            //std::cout << "Compute view image" << std::endl;
            place.draw_image_portion (img_col, img_res2);
            if (click_n > 0) {
                if (place.zoom < 10)
                     place.draw_cell_cross (img_res2, cell_x, cell_y);
                else place.draw_cell_rect  (img_res2, cell_x, cell_y);
            }
            std::string s1 = edit_mode ? "edit" : "",
                        s2 = click_n > 0 ? compose_msg_map_coords () : "",
                        s3 = s1 != "" && s2 != "" ? " " : "";
            set_right_footer (s1+s3+s2);
            place.draw_footer (img_res2);
            cv::imshow ("View", img_res2);
        }

        recomp = R_NOTHING;

        // Wait next event on all windows, with a 15ms timeout to detect
        // flag changes
        int key = cv::waitKey (15);

        // Manage keyboard events using a "handmade" callback that we call
        // ourself. The mouse and slider callbacks are called by waitkey() during
        // the waiting.
        if (onKeyPressEvent (key, this) < 0) break;
    }
}


// Callbacks of sliders

void Astico2D::onZoomSlide (int pos, void *data)
{
    Astico2D *ao = (Astico2D*) data;
    ao->place.zoom = (ao->zoom_slider > 0) ? ao->zoom_slider : 1;
    ao->place.adjust_bounds ();
    ao->raise_recomp (R_MAP);
}

void Astico2D::onDefaultTrackbarSlide (int pos, void *data)
{
    Astico2D *ao = (Astico2D*) data;
    ao->raise_recomp (R_ALL);
}


// Callbacks of mouse

void Astico2D::onMouseEventMap (int event, int x, int y, int flags, void *data)
{
    Astico2D *ao = (Astico2D*) data;
    switch (event) {
        case cv::EVENT_LBUTTONDOWN :
            ao->click_x = x; ao->click_dx = 0;
            ao->click_y = y; ao->click_dy = 0;
            ao->click_n = 1;
            break;
        case cv::EVENT_MOUSEMOVE :
            // std::cout << "mouse move " << x << "," << y << std::endl;
            if (ao->click_n == 1) {
                int dx = (x - ao->click_x) / ao->place.map_f, 
                    dy = (y - ao->click_y) / ao->place.map_f;
                if (ao->click_dx != dx || ao->click_dy != dy) {
                    ao->place.move (dx - ao->click_dx, dy - ao->click_dy);
                    ao->click_dx = dx;
                    ao->click_dy = dy;
                    ao->raise_recomp (R_MAP);
                }
            }
            break;
        case cv::EVENT_LBUTTONUP :
            ao->click_n = 0;
            break;
    }
}


void Astico2D::onMouseEventView (int event, int x, int y, int flags, void *data)
{
    Astico2D *ao = (Astico2D*) data;
    bool memo_cell = false, 
         flag_ctrl = flags & cv::EVENT_FLAG_CTRLKEY;

    switch (event) {
        case cv::EVENT_LBUTTONDOWN :
            ao->click_x = x;
            ao->click_y = y;
            ao->click_n = 1;
            memo_cell = true;
            break;
        case cv::EVENT_MBUTTONDOWN :
            ao->click_x = x; ao->click_dx = 0;
            ao->click_y = y; ao->click_dy = 0;
            ao->click_n = 2;
            memo_cell = true;
            break;
        case cv::EVENT_RBUTTONDOWN :
            ao->click_x = x;
            ao->click_y = y;
            ao->click_n = 3;
            if (ao->mode_transform == M_LEVELS)
                 ao->place.print_array_values (ao->img_int, x, y, 
                     flag_ctrl ? 10 : 5, flag_ctrl ? 10 : 4);
            else ao->place.print_array_colors (ao->img_col, x, y, 
                     flag_ctrl ? 6 : 2, flag_ctrl ? 6 : 3);
            memo_cell = true;
            break;

        case cv::EVENT_LBUTTONUP :
        case cv::EVENT_MBUTTONUP :
        case cv::EVENT_RBUTTONUP :
            ao->click_n = 0;
            ao->cell_x = ao->cell_y = -1;
            ao->raise_recomp (R_VIEW);
            break;

        case cv::EVENT_MOUSEMOVE :
            // std::cout << "mouse move " << x << "," << y << std::endl;
            if (ao->click_n == 1) {
                memo_cell = true;
            } else if (ao->click_n == 2) {
                int dx = (x - ao->click_x) / ao->place.zoom, 
                    dy = (y - ao->click_y) / ao->place.zoom;
                if (ao->click_dx != dx || ao->click_dy != dy) {
                    ao->place.move (ao->click_dx - dx, ao->click_dy - dy);
                    ao->click_dx = dx;
                    ao->click_dy = dy;
                    ao->raise_recomp (R_MAP);
                }
            }
            break;
    }

    bool cell_changed = false;
    if (memo_cell) {
        int old_cell_x = ao->cell_x, old_cell_y = ao->cell_y;
        ao->place.get_cell_coords (x, y, ao->cell_x, ao->cell_y);
        cell_changed = old_cell_x != ao->cell_x || old_cell_y != ao->cell_y;
        ao->raise_recomp (R_VIEW);
    }
    if (ao->edit_mode) {
        if (ao->click_n == 1) {
            if ( event == cv::EVENT_LBUTTONDOWN ||
                (event == cv::EVENT_MOUSEMOVE && cell_changed) ) {
                reverse_pixel_colors (ao->img_src, ao->cell_x, ao->cell_y);
                ao->raise_recomp (R_ALL);
            }
        }
    }
}


// Keyboard management

void Astico2D::print_keyboard_keys ()
{
    std::cout <<
        "Mouse button in the View window:\n"
        "  left     drag mouse to see values in the footer\n"
        "  middle   drag mouse to move the view\n"
        "  right    print values/colors in the terminal, ctrl to enlarge\n"
        "Keyboard:\n"
        "  a     show this help\n"
        "  Esc   quit without saving\n"
        "  hHlL  change width or height of the View\n"
        "  jJkK  move the view\n"
        "  zZ    change zoom\n"
        "  g     maximize View and adapt the zoom\n"
        "  mM    change width of the Map\n"
        "  ou    show original or thresholded image\n"
        "  i     reverse colors of original image\n"
        "  tT    change threshold\n"
        "  e     toggle edit mode (left button to reverse pixel colors)\n"
        "  r     reload input image\n"
        "  w     save output image\n";
}


int Astico2D::onKeyPressEvent (int key, void *data)
{
    Astico2D *ao = (Astico2D*) data;
    if (key < 0) return 0;        // no key pressed
    key &= 255;                   // to compare with a char

    switch (key) {
        case 'a' :
            ao->print_keyboard_keys();
            break;
        case 27 : // ESC to quit
            return -1;
        case 'h' :
        case 'H' :
        case 'l' :
        case 'L' : {
            int dir_h = (key == 'h') ? -1 : (key == 'H') ? 1 : 0;
            int dir_w = (key == 'l') ? -1 : (key == 'L') ? 1 : 0;
            ao->place.resize_view (dir_h, dir_w);
            ao->img_res2 = cv::Mat(ao->place.zoom_h + ao->place.footer_h, 
                ao->place.zoom_w, CV_8UC3);
            ao->raise_recomp (R_MAP);
            break;
        }
        case 'j' :
            ao->place.move (0, -1);
            ao->raise_recomp (R_MAP);
            break;
        case 'J' :
            ao->place.move (0, 1);
            ao->raise_recomp (R_MAP);
            break;
        case 'k' :
            ao->place.move (-1, 0);
            ao->raise_recomp (R_MAP);
            break;
        case 'K' :
            ao->place.move (1, 0);
            ao->raise_recomp (R_MAP);
            break;
        case 'z' : 
            ao->decrease_slider ("Zoom", 1);
            break;
        case 'Z' : 
            ao->increase_slider ("Zoom", ao->place.zoom_max);
            break;
        case 'g' : {
            int z = ao->place.compute_optimal_zoom ();
            ao->modify_slider ("Zoom", z);
            ao->place.resize_view_opt ();
            ao->img_res2 = cv::Mat(ao->place.zoom_h + ao->place.footer_h, 
                ao->place.zoom_w, CV_8UC3);
            ao->raise_recomp (R_MAP);
            break;
        }
        case 'm' :
            ao->place.change_map_width (-1);
            ao->raise_recomp (R_MAP);
            break;
        case 'M' :
            ao->place.change_map_width (1);
            ao->raise_recomp (R_MAP);
            break;
        case 'o' :
            ao->set_left_footer ("Original image");
            ao->raise_recomp (R_ALL);
            ao->mode_transform = M_NONE;
            ao->key_transform = 0;
            break;
        case 'i' :
            ao->set_left_footer ("Colors reversed");
            reverse_img_colors (ao->img_src);
            ao->raise_recomp (R_ALL);
            break;
        case 'u' :
            ao->set_left_footer ("Thresholded image");
            ao->raise_recomp (R_ALL);
            ao->mode_transform = M_LEVELS;
            ao->key_transform = 0;
            break;
        case 't' :
            ao->decrease_slider ("Thresh", 0); 
            break;
        case 'T' : 
            ao->increase_slider ("Thresh", ao->thresh_max);
            break;
        case 'e' :
            ao->edit_mode = ! ao->edit_mode;
            ao->raise_recomp (R_VIEW);
            break;
        case 'r' :
            if (ao->read_input_image () == 0) {
                 ao->set_left_footer ("Image reloaded");
                 ao->raise_recomp (R_ALL);
            } else ao->set_left_footer ("Read error");
            break;
        case 'w' :
            if (ao->save_result_image () == 0)
                 ao->set_left_footer ("Image saved");
            else ao->set_left_footer ("Write error");
            break;

        default :
            if (ao->handle_keyboard (key)) {
                ao->raise_recomp (R_ALL);
            }
            break;
    }
    return 1;
}


//------------------------ M E A S U R E   T I M E ----------------------------

// Measure elapsed time.
// Code extracted from EZ-Draw version 1.2 (LGPL), Edouard.Thiel@univ-amu.fr

#if MEASURE_TIME

AoMeasureTime::AoMeasureTime ()
{
#  ifdef _WIN32
_init_timeofday ();
#  endif
}


// Return the elapsed time since the Epoch in seconds,microseconds
double AoMeasureTime::get_time_dbl ()
{
    struct timeval tp;
    _gettimeofday (&tp);
    return (double) tp.tv_sec + (double) tp.tv_usec / (double) 1E6;
}


#  ifdef _WIN32

// Initialize time computation.

void AoMeasureTime::_init_timeofday (void)
{
    LARGE_INTEGER freq;

    _perf_freq = 0;
    _get_systemtime (&_start_time);
    QueryPerformanceCounter (&_start_count);

    if (QueryPerformanceFrequency (&freq) && freq.QuadPart > 0)
         _perf_freq = 1.0 / freq.QuadPart;
}


// Emulate gettimeofday() with GetSystemTimeAsFileTime().
// GetSystemTimeAsFileTime() has an accuracy of 15ms (10ms on XP).
// GetSystemTimePreciseAsFileTime() has an accuracy of 1ms but needs
// windows >= 8, and seems rather expensive.

void AoMeasureTime::_get_systemtime (struct timeval *t)
{
    ULARGE_INTEGER ul;
    FILETIME ft;
    double dt;
    if (t == NULL) return;
    GetSystemTimeAsFileTime (&ft);
    ul.LowPart = ft.dwLowDateTime;
    ul.HighPart = ft.dwHighDateTime;
    dt = ul.QuadPart / 10000000.0 - 11644473600.0;
    t->tv_sec = dt;
    t->tv_usec = (dt - t->tv_sec) * 1000000.0;
}


// Emulate gettimeofday() with QueryPerformanceCounter().
// This method seems to be more accurate than _get_systemtime()
// but leads to a drift.

void AoMeasureTime::_get_countertime (struct timeval *t)
{
    LARGE_INTEGER now;
    double dt;
    QueryPerformanceCounter (&now);
    dt = (now.QuadPart - _start_count.QuadPart) * _perf_freq;
    *t = _start_time;
    t->tv_sec += dt;
    t->tv_usec += (dt - (int)dt) * 1000000;
    t->tv_sec += t->tv_usec / 1000000;
    t->tv_usec %= 1000000;
}

#  endif // _WIN32


// Compute elapsed time since the Epoch.

void AoMeasureTime::_gettimeofday (struct timeval *t)
{
    if (t == NULL) return;
#ifdef _WIN32
    if (_perf_freq == 0)
         _get_systemtime (t);
    else _get_countertime (t);
#else
    gettimeofday (t, NULL);
#endif
}

#endif // MEASURE_TIME

